using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Drawing.Imaging;

namespace TextSharp {
    public class TextSharp {

        /// <summary>
        /// Graphics object used to use GDI+ drawing on the console.
        /// </summary>
        public readonly Graphics Graphics;

        /// <summary>
        /// Bitmap used to store the current work surface
        /// </summary>
        private Bitmap workSurface;
        
        /// <summary>
        /// Converts 16bpp values to the required data on the console
        /// </summary>
        private CharInfo[] rgbConversionTable;
        /// <summary>
        /// Handle used for writing output.
        /// </summary>
        private IntPtr outputHandle;
        /// <summary>
        /// Internal buffer to store characters before they are copied to the console.
        /// </summary>
        private CharInfo[] characterBuffer;

        /// <summary>
        /// Size of the buffer (width and height)
        /// </summary>
        private COORD bufferSize;
        /// <summary>
        /// Position of the buffer for writing (0,0)
        /// </summary>
        private COORD bufferPosition;
        /// <summary>
        /// Area to write to the buffer (inclusive)
        /// </summary>
        private SMALL_RECT writeArea;
        /// <summary>
        /// Area used when bitmap is locked for reading (exclusive)
        /// </summary>
        private Rectangle fullBitmapRect;

        /// <summary>
        /// Width of the console window and buffer (in characters).
        /// </summary>
        public readonly int Width;
        /// <summary>
        /// Height of the console window and buffer (in characters).
        /// </summary>
        public readonly int Height;

        /// <summary>
        /// Gets or sets the title to display in the console title bar.
        /// </summary>
        public string Title {
            get { return Console.Title; }
            set { Console.Title = value; }
        }

        /// <summary>
        /// Set up the console for use with the TextSharp class.
        /// </summary>
        /// <param name="width">Desired width of the console in characters.</param>
        /// <param name="height">Desired height of the console in characters.</param>
        public TextSharp(int width, int height) {

            this.Width = width;
            this.Height = height;

            this.workSurface = new Bitmap(width, height, PixelFormat.Format16bppRgb565);
            this.Graphics = Graphics.FromImage(workSurface);
            
            Console.SetBufferSize(width, height);
            Console.SetWindowSize(width, height);
            
            this.characterBuffer = new CharInfo[width * height];


            this.fullBitmapRect = new Rectangle(0, 0, width, height);
            this.bufferSize = new COORD((ushort)width, (ushort)height);
            this.bufferPosition = new COORD(0, 0);
            this.writeArea = new SMALL_RECT(0, 0, (ushort)(width - 1), (ushort)(height - 1));

            this.rgbConversionTable = new CharInfo[0x10000];

            byte[] palette = Properties.Resources.ConsolePalette;
            for (int i = 0; i < palette.Length; i += 2) {
                this.rgbConversionTable[i >> 1] = new CharInfo(palette[i + 0], palette[i + 1]);
            }
            outputHandle = GetStdHandle(StandardDevice.Output);
        }

        /// <summary>
        /// Character information structure
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct CharInfo {
            internal char character;

            public char Character {
                get { return this.character; }
                set { this.character = value; }
            }

            /// <summary>
            /// Colour attributes of the character
            /// </summary>
            internal ushort Attributes;

            /// <summary>
            /// Gets or sets the foreground colour of the character
            /// </summary>
            public ConsoleColor ForegroundColour {
                get { return (ConsoleColor)(this.Attributes & 0x0F); }
                set { this.Attributes = (ushort)((this.Attributes & 0xF0) | ((int)value & 0x0F)); }
            }

            /// <summary>
            /// Gets or sets the background colour of the character
            /// </summary>
            public ConsoleColor BackgroundColour {
                get { return (ConsoleColor)(this.Attributes >> 4); }
                set { this.Attributes = (ushort)((this.Attributes & 0x0F) | ((int)value << 4)); }
            }


            public CharInfo(byte c, byte a) {
                this.Attributes = (ushort)a;
                this.character = (char)c;                
            }

            public override string ToString() {
                return string.Format("{0} {{{1}, {2}}}", this.character, this.ForegroundColour, this.BackgroundColour);
            }

        }

        /// <summary>
        /// Coordinate structure
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        private struct COORD {
            internal ushort x;
            internal ushort y;
            public COORD(ushort x, ushort y) {
                this.x = x;
                this.y = y;
            }
        }

        /// <summary>
        /// Update the console window with the contents of the image
        /// </summary>
        public void Refresh() {
            unsafe {
                BitmapData bd = this.workSurface.LockBits(fullBitmapRect, ImageLockMode.ReadOnly, PixelFormat.Format16bppRgb565);
                ushort* p = (ushort*)bd.Scan0.ToPointer();
                for (int i = 0; i < Width * Height; ++i) {
                    this.characterBuffer[i] = this.rgbConversionTable[*p++];
                }
                this.workSurface.UnlockBits(bd);
                fixed (CharInfo* q = this.characterBuffer) {
                    bool j = WriteConsoleOutput(outputHandle, (IntPtr)q, bufferSize, bufferPosition, ref writeArea);
                }
            }
        }

        /// <summary>
        /// Rectangle structure
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        private struct SMALL_RECT {
            internal ushort Left;
            internal ushort Top;
            internal ushort Right;
            internal ushort Bottom;
            public SMALL_RECT(ushort left, ushort top, ushort right, ushort bottom) {
                this.Left = left;
                this.Top = top;
                this.Right = right;
                this.Bottom = bottom;
            }
        }

        /// <summary>
        /// Standard device handle identifier
        /// </summary>
        private enum StandardDevice : int {
            Input = -10,
            Output = -11,
            Error = -12
        };

        /// <summary>
        /// Retrieve a handle for a standard device
        /// </summary>
        /// <param name="nStdHandle">Handle indentifier</param>
        /// <returns>The handle</returns>
        [DllImport("kernel32.dll")]
        private static extern IntPtr GetStdHandle(StandardDevice nStdHandle);

        /// <summary>
        /// Copy an array of CHAR_INFO structures to a console window.
        /// </summary>
        /// <param name="handle">Output handle of the console window you wish to write to.</param>
        /// <param name="buffer">Pointer to the buffer you wish to copy.</param>
        /// <param name="bufferSize">Dimensions of the buffer you are copying.</param>
        /// <param name="bufferPosition">Position on the console window you wish to copy to.</param>
        /// <param name="writeRegion">Region that you are writing to.</param>
        /// <returns></returns>
        [DllImport("kernel32.dll", EntryPoint = "WriteConsoleOutputA")]
        private static extern bool WriteConsoleOutput(IntPtr handle, IntPtr buffer, COORD bufferSize, COORD bufferPosition, ref SMALL_RECT writeRegion);
    }
}
